/* Device Driver (Simulation) for XADIO Cards  */
/* Copyright (1999-2008) Brueckner & Jarosch Erfurt Germany */

#include <linux/module.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/sched.h>
#include <linux/fcntl.h>
#include <linux/delay.h>
#include <linux/proc_fs.h>
#include <linux/pci.h>
#include <linux/poll.h>
#include <linux/ctype.h>

#define XADIO_NO                          4
#define XADIO_DRIVERVERSION               64
#define XADIO_DEFAULT_BRECORDS            1024
#define XADIO_DEFAULT_BCOUNT              8
#define XADIO_DEFAULT_NOSUBVENDORCHECK    0
#define XADIO_DEFAULT_HWVERSION           2

#define ARRSIZE(a)          ((int)(sizeof(a)/sizeof(a[0])))
#define OFFSET(type,member) ((int)(&((type*)0)->member))
#define ARRPOS(ptr,arr)     ((ptr)-(arr))
#define MIN(a,b)            (((a)<(b))?(a):(b))
#define MAX(a,b)            (((a)>(b))?(a):(b))
#define LO(ll)              ((unsigned int)((ll      )&0xffffffffLL))
#define HI(ll)              ((unsigned int)(((ll)>>32)&0xffffffffLL))
#define HILO(ll)            HI(ll),LO(ll)

#include "xadio.h"

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
#undef MOD_INC_USE_COUNT
#define MOD_INC_USE_COUNT
#undef MOD_DEC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

#ifndef MODULE_LICENSE
#define MODULE_LICENSE(x)
#endif

typedef struct {
  int flags;
  char name[10];
  int rate;
  int incmode[2];
  int dac[2];
  int digio;
  char config_rom[XADIO_CONFIG_ROMSIZE];
} xadio_t;

// Module Parameter
static int debug=0;
static int major = 0x46; 
static int brecords = XADIO_DEFAULT_BRECORDS;
static int bcount = XADIO_DEFAULT_BCOUNT;

// Globals
static int bsize = 0;
static xadio_t xadio_table[2];
#define XADIO_NAME(minor)        (xadio_table[minor].name)

static void xadio_sleep(int minor,int msec) {
wait_queue_head_t wqh;
  init_waitqueue_head(&wqh);
  interruptible_sleep_on_timeout(&wqh,(msec*HZ)/1000);
}

static unsigned long long xadio_hwtimer(int minor) {
  long long val;
  rdtscll(val);
  return val;
}

// Returns CPU-Clocks / 128 
// example 2Ghz = 2000000000/12500000 = 160 would be good
// now            2000000000/128=15625000 ca 16Mhz

static loff_t xadio_currpos(int minor) {
unsigned long long pos=xadio_hwtimer(minor);
  return (pos&0x7fffffffffffffffLL)>>7; //depends on cpu 
}

static inline int xadio_readok(int minor,loff_t pos) {
loff_t maxpos=xadio_currpos(minor)-bsize; // now
loff_t minpos=maxpos-(bcount-1)*bsize; // would bee so
  if(pos<minpos) return -1; // overflow
  if(pos>maxpos) return 1;
  return 0;
}

/* Device-IO Functions */

static ssize_t xadio_write(struct file * file, const char * buf, size_t count, loff_t *ppos) {
unsigned int minor=MINOR(file->f_dentry->d_inode->i_rdev);
  if(debug) printk(KERN_DEBUG "%s write: count:%d\n",XADIO_NAME(minor),(int)count);
  return count;
}

static loff_t xadio_lseek(struct file * file, loff_t offset, int orig) {
unsigned int minor=MINOR(file->f_dentry->d_inode->i_rdev);
loff_t result;
  if(debug>1) printk(KERN_DEBUG "%s lseek: %d %d\n",XADIO_NAME(minor),(int)offset,orig);
  switch (orig) {
    case 0:
      result=(file->f_pos=offset);
      break;
    case 1:
      result=(file->f_pos+=offset);
      break;
    case 2:
      result=(file->f_pos=xadio_currpos(minor)+offset);
      break;
    default: 
      result=-EINVAL;
      break;
  }
  if(debug>1) printk(KERN_DEBUG "%s lseek(result): %8.8x%8.8x\n",XADIO_NAME(minor),HILO(result));
  return result;
}

static int xadio_open(struct inode * inode, struct file * file) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
int minor=MINOR(inode->i_rdev);
#else
struct proc_dir_entry *entry=PDE(inode);
int minor=(int)(entry->data);
  inode->i_rdev=MKDEV(major,minor);
#endif
  if(minor>1) return -ENODEV;
  if(debug) printk(KERN_DEBUG "%s open\n", XADIO_NAME(minor));
  MOD_INC_USE_COUNT;
  return 0;
}

static int xadio_release(struct inode * inode, struct file * file) {
unsigned int minor = MINOR(inode->i_rdev);
  if(debug) printk(KERN_DEBUG "%s release\n", XADIO_NAME(minor));
  MOD_DEC_USE_COUNT;
  return 0;
}

static unsigned int xadio_poll(struct file *file,poll_table *wait) {
unsigned int minor=MINOR(file->f_dentry->d_inode->i_rdev);
int result=0;
  if(debug>1) printk(KERN_DEBUG "%s poll\n", XADIO_NAME(minor));
  switch(xadio_readok(minor,file->f_pos)) {
    case -1:
    case 0:
      result=POLLIN | POLLRDNORM;
      break;
  }
  if(debug>1) printk(KERN_DEBUG "%s poll result:%d\n", XADIO_NAME(minor),result);
  return result;
}

static int simultimer32(int minor,loff_t pos) {
  return (pos>>6)&0x7fffffff;
}

static int simulinc0(int minor,loff_t pos) {
  return (simultimer32(minor,pos)>>2)&0x3ff;
}

static int simulinc1(int minor,loff_t pos) {
  return (-(simultimer32(minor,pos)>>3))&0x3ff;
}

static int simuladu(int minor,int channel,loff_t pos) {
//loff_t sample=simultimer32(minor,pos);
int angle0=simulinc0(minor,pos);
//int angle1=simulinc1(minor,pos);
  switch(channel) {
    case 0:
      return angle0+0x8000;
    case 1:
      return 0x8000-angle0;
    case 2:
      return 0x8000-4*(angle0%256);
    case 3:
      return angle0;
    case 4:
      return 0x8000+minor;
    case 5:
      return 0xffff-angle0;
    case 6:
      return 1024-angle0;
    case 7:
      return 0x8000+4*(angle0%256);
  }
  return 0x33333333;
}

static int simuldigio(int minor,loff_t pos) {
int angle0=simulinc0(minor,pos);
int angle1=simulinc1(minor,pos);
int result=xadio_table[minor].digio;
  result &= ~(XADIO_COUNTER0_RESET | XADIO_COUNTER1_RESET);
  if(angle0==0) result|=XADIO_COUNTER0_RESET;
  if(angle1==0) result|=XADIO_COUNTER1_RESET;
  return result;
}

static void gensimul(int minor,Xadio_record *s,loff_t pos) {
  s->timer32=simultimer32(minor,pos);
  s->adc[0]=simuladu(minor,0,pos);
  s->adc[1]=simuladu(minor,1,pos);
  s->adc[2]=simuladu(minor,2,pos);
  s->adc[3]=simuladu(minor,3,pos);
  s->adc[4]=simuladu(minor,4,pos);
  s->adc[5]=simuladu(minor,5,pos);
  s->adc[6]=simuladu(minor,6,pos);
  s->adc[7]=simuladu(minor,7,pos);
  s->digio=simuldigio(minor,pos);
  s->inc0=simulinc0(minor,pos);
  s->inc1=simulinc1(minor,pos);
  s->dac0=xadio_table[minor].dac[0];
  s->dac1=xadio_table[minor].dac[1];
  s->fi1=0xadda020a;
  s->fi2=0xf0000000;
}

static ssize_t xadio_read(struct file * file, char * buf, size_t count, loff_t *ppos) {
unsigned int minor=MINOR(file->f_dentry->d_inode->i_rdev);
loff_t pos=*ppos;
size_t done=0;
  if(debug>1) printk(KERN_DEBUG "%s read: off:0x%8.8x%8.8x count:%d\n",XADIO_NAME(minor),HILO(pos),count);
  if(count%sizeof(Xadio_record)) done=-EINVAL;
  while(!done) {
    switch(xadio_readok(minor,pos)) {
      case -1: {
        // overrun 
        loff_t curr=xadio_currpos(minor);
        if(debug>1) printk(KERN_WARNING "%s overrun,skipping:0x%8.8x%8.8x starting 0x%8.8x%8.8x from 0x%8.8x%8.8x\n",XADIO_NAME(minor),HILO(curr-pos),HILO(pos),HILO(curr));
        pos=curr;
        }
      case 0: { // normal
        int i; 
        Xadio_record simulbuffer;
        for(i=0;i<(int)(count/sizeof(Xadio_record));i++) {
          gensimul(minor,&simulbuffer,pos);
          copy_to_user(buf+done,&simulbuffer,sizeof(Xadio_record));
          done+=sizeof(Xadio_record);
          pos+=sizeof(Xadio_record);
        }
        break;
      }
      case 1: //wait
        xadio_sleep(minor,10);
        if(signal_pending(current)) done=-EINTR;
        break;
    }
  } 
  *ppos=pos;
  if(debug>1) printk(KERN_DEBUG "%s read: result:%d\n",XADIO_NAME(minor),(int)done);
  return done;
}

static struct file_operations xadio_fops = {
  owner:  THIS_MODULE,
  llseek: xadio_lseek,
  read:   xadio_read,
  write:  xadio_write,
  poll:   xadio_poll,
  open:   xadio_open,
  release:  xadio_release
};

//------- procfs code -----------------------------------------

static int xadio_getrate(int minor) {
  return xadio_table[minor].rate;
}

static const int xadio_quarz=50000000;

static int xadio_base(void) {
  return xadio_quarz/128;
}

static int xadio_rate2freq(int rate) {
  return xadio_base()/(rate+1);
}

static int xadio_freq2rate(int freq) {
  return (xadio_base()/freq)-1;
}

static int xadio_getfreq(int minor,int channel) {
  return xadio_rate2freq(xadio_getrate(minor));
}

static void xadio_setrate(int minor,int rate) {
  if(debug) printk(KERN_DEBUG "%s xadio_setrate %d\n",XADIO_NAME(minor),rate);
  xadio_table[minor].rate=rate;
}

static int xadio_setfreq(int minor,int freq) {
int rate;
  if(freq<5000) freq=5000;
  if(freq>200000) freq=200000;
  rate=xadio_freq2rate(freq);
  xadio_setrate(minor,rate);
  return 0;
}

static void xadio_setincmode0(int minor,int mode) {
  xadio_table[minor].incmode[0]=mode;
}

static void xadio_setincmode1(int minor,int mode) {
  xadio_table[minor].incmode[1]=mode;
}

static int xadio_getincmode(int minor,int channel) {
  if(minor>1 || channel>1 || channel<0 || minor<0) return 0;
  return xadio_table[minor].incmode[channel];
}

/* Fixed to +-10.24V */
static int xadio_getchanneloffset(int minor,int channel) {
  return 10240000;
}

/* returns gain * 100  because of precision */
static int xadio_getchannelgain100(int minor,int channel) {
  return 31250;
}

#ifdef LATER
static int xadio_getadc(int minor,int channel) {
  return GetFpga_adu(minor,channel);
}
#endif

static int xadio_adc2value(int adc) {
 return (((adc) & 0xffff0000)? 0 : ((adc)?(adc):1));
}

static int xadio_adc2error(int adc) {
  return ((adc)>>16)&0x3;
}

static int xadio_uv2dac(int minor,int channel,int uv) {
  return (((uv<<3)/625)&0x3ffff);
}

static int xadio_dac2uv(int minor,int channel,int dac) {
int result=(dac&0x20000)?((int)(dac|0xfffc0000)):dac;
  return (result*625)>>3;
}

static int xadio_value2uv(int minor,int channel,int value) {
int offset=xadio_getchanneloffset(minor,channel);
int gain=xadio_getchannelgain100(minor,channel);
  return (value*gain)/100-offset;
}

static void xadio_setdac(int minor,int channel,int value) {
  channel&=1;
  xadio_table[minor].dac[channel]=value;
}

static void xadio_setdacuv(int minor,int channel,int value) {
  xadio_setdac(minor,channel,xadio_uv2dac(minor,channel,value));
}

static void xadio_setdigio(int minor,u32 value,u32 mask) { 
u32 orig=xadio_table[minor].digio & ~mask;
  if(debug>1) printk(KERN_DEBUG "%s xadio_setdigio Value: %x Mask: %x\n",XADIO_NAME(minor),value,mask);
  xadio_table[minor].digio=orig|(value&mask);
}

typedef int (*GetFkt)(int minor,int channel);
typedef int (*SetFkt)(int minor,int channel,int value);

static int xadio_getversion(int minor,int channel) { return XADIO_DRIVERVERSION; }
static int xadio_getfpgaversion(int minor,int channel) { return 0xadda002a; }
static int xadio_gethwversion(int minor,int channel) { return 2; }
static int xadio_getbrecords(int minor,int channel) { return brecords; }
static int xadio_getbcount(int minor,int channel) { return bcount; }
static int xadio_getadcbits(int minor,int channel) { return 16; }
static int xadio_getdacbits(int minor,int channel) { return 18; }
static int xadio_getchannels(int minor,int channel) { return 8; }

static int xadio_gettimer32(int minor,int channel) { return simultimer32(minor,xadio_currpos(minor)); }
static int xadio_getinc0(int minor,int channel) { return simulinc0(minor,xadio_currpos(minor)); }
static int xadio_getinc1(int minor,int channel) { return simulinc1(minor,xadio_currpos(minor)); }
static int xadio_getdigio(int minor,int channel) { return simuldigio(minor,xadio_currpos(minor)); }

typedef struct {
  int mode;
  int channel;
  const char *name;
  GetFkt   fkt;
} FormatEntry;

static FormatEntry controlArray[] = {
  { 0, 0, XADIO_CONTROL_version,     xadio_getversion },
  { 0, 0, XADIO_CONTROL_hwversion ,  xadio_gethwversion },
  { 1, 0, XADIO_CONTROL_fpgaversion, xadio_getfpgaversion   },
  { 0, 0, XADIO_CONTROL_brecords ,   xadio_getbrecords },
  { 0, 0, XADIO_CONTROL_bcount ,     xadio_getbcount },
  { 0, 0, XADIO_CONTROL_freq ,       xadio_getfreq },
  { 0, 0, XADIO_CONTROL_incmode0,    xadio_getincmode },
  { 0, 1, XADIO_CONTROL_incmode1,    xadio_getincmode },
  { 0, 0, XADIO_CONTROL_adcbits,     xadio_getadcbits },
  { 0, 0, XADIO_CONTROL_dacbits,     xadio_getdacbits },
  { 0, 0, XADIO_CONTROL_channels,    xadio_getchannels },
  { 6, 0, XADIO_CONTROL_frequencies, (GetFkt)0 },
  { 7, 0, XADIO_CONTROL_configrom,   (GetFkt)0 }
};

static FormatEntry valuesArray[] = {
  { 2, 0, XADIO_VALUES_timer,       (GetFkt)0         },
  { 0, 0, XADIO_VALUES_timer32,     xadio_gettimer32  },
  { 0, 0, XADIO_VALUES_inc0,        xadio_getinc0     },
  { 0, 0, XADIO_VALUES_inc1,        xadio_getinc1     },
  { 4, 0, XADIO_VALUES_adc0 ,       (GetFkt)0         },
  { 4, 1, XADIO_VALUES_adc1 ,       (GetFkt)0         },
  { 4, 2, XADIO_VALUES_adc2 ,       (GetFkt)0         },
  { 4, 3, XADIO_VALUES_adc3 ,       (GetFkt)0         },
  { 4, 4, XADIO_VALUES_adc4 ,       (GetFkt)0         },
  { 4, 5, XADIO_VALUES_adc5 ,       (GetFkt)0         },
  { 4, 6, XADIO_VALUES_adc6 ,       (GetFkt)0         },
  { 4, 7, XADIO_VALUES_adc7 ,       (GetFkt)0         },
  { 5, 0, XADIO_VALUES_dac0,        (GetFkt)0         },
  { 5, 1, XADIO_VALUES_dac1,        (GetFkt)0         },
  { 3, 0, XADIO_VALUES_digio,       xadio_getdigio    }
};

static int xadio_bool(unsigned int value,int bits,char *buffer,int size) {
int i;
int pos=0;
  if(size<bits+bits/4+2) return 0;
  for(i=0;i<bits;i++) {
    if(!(i%4)) buffer[pos++]='.';
    buffer[pos++]=(value&(0x1<<(bits-i-1)))?'1':'0';
  }
  buffer[pos++]=0;
  return pos;
}

static int xadio_format(int minor,char *buffer,int size,FormatEntry *fe,int asize) {
int count=0;
int i;
  for(i=0;i<asize;i++) {
    if(count>size-50) {
    	break;
    }	
    switch(fe[i].mode) {
      case 0:
        count+=sprintf(buffer+count,"%s %d\n",fe[i].name,(fe[i].fkt)(minor,fe[i].channel));
        break;
      case 1:
        count+=sprintf(buffer+count,"%s 0x%x\n",fe[i].name,(fe[i].fkt)(minor,fe[i].channel));
        break;
      case 2: {
        loff_t currpos=xadio_currpos(minor);
        count+=sprintf(buffer+count,"%s 0x%8.8x%8.8x\n",fe[i].name,HILO(currpos));
        }
        break;
      case 3: {
        char lbuffer[100];
        int value=(fe[i].fkt)(minor,fe[i].channel);
        xadio_bool(value,24,lbuffer,sizeof(lbuffer));
        count+=sprintf(buffer+count,"%s 0x%6.6x 0b%s\n",fe[i].name,value,lbuffer);
        break;
      }
      case 4: {
        int channel=fe[i].channel;
        int adc=simuladu(minor,channel,xadio_currpos(minor));
        int value=xadio_adc2value(adc);
        int uv=xadio_value2uv(minor,channel,value);
        int error=xadio_adc2error(adc);
        count+=sprintf(buffer+count,"%s 0x%6.6x % 9d %d\n",fe[i].name,value,uv,error);
        break;
      }
      case 5: {
        int channel=fe[i].channel;
        int dac=xadio_table[minor].dac[channel];
        int uv=xadio_dac2uv(minor,channel,dac);
        count+=sprintf(buffer+count,"%s 0x%6.6x % 9d\n",fe[i].name,dac,uv);
        break;
      }
      case 6: {
        int j;
        count+=sprintf(buffer+count,"%s {",fe[i].name);
        for(j=0;j<128;j++) {
          int freq=xadio_rate2freq(j);
          if(freq>200000 || freq<5000) continue; // AD7222 Anschlag
          if(count>size-50) continue;
          count+=sprintf(buffer+count,"%d ",freq);
        }
        count+=sprintf(buffer+count,"}\n");
        break;
      }
      case 7: {
        int j; 
        count+=sprintf(buffer+count,"%s ",fe[i].name);
        for(j=0;j<XADIO_CONFIG_ROMSIZE;j++) {
          if(count>size-50) continue;
          count+=sprintf(buffer+count,"%02x",((unsigned char*)(xadio_table[minor].config_rom))[j]);
        }
        count+=sprintf(buffer+count,"\n");
        break;
      }
    }
  }
  buffer[count]=0;
  return count;
}

static int xadio_common_format(int minor,char *page, char **start, off_t off, int count, int *eof, void *data,FormatEntry *fe,int asize) {
int size;
  if(off!=0) { *eof=1; return 0; }
  size=xadio_format(minor,page,count,fe,asize);
  *eof=1;
  *start=page;
  return size;
}

static int xadio_control_read(char *page, char **start, off_t off, int count, int *eof, void *data) {
int minor=((unsigned int)data) & 0xff;
  if(debug) printk(KERN_DEBUG "%s xadio_control_read: off:%d count:%d\n",XADIO_NAME(minor),(int)off,(int)count);
  return xadio_common_format(minor,page,start,off,count,eof,data,controlArray,ARRSIZE(controlArray));
}

static int xadio_values_read(char *page, char **start, off_t off, int count, int *eof, void *data) {
int minor=((unsigned int)data) & 0xff;
  if(debug) printk(KERN_DEBUG "%s xadio_values_read: off:%d count:%d\n",XADIO_NAME(minor),(int)off,(int)count);
  return xadio_common_format(minor,page,start,off,count,eof,data,valuesArray,ARRSIZE(valuesArray));
}


static int xadio_scanint(const char *buffer,int len,int *result) {
char lbuffer[20];
char *endptr;
int sign=1;
  if(len>0 && buffer[0]=='-') {
    sign=-1;
    buffer++;
    len--;
  }
  len=MIN(len,(int)sizeof(lbuffer)-1);
  memcpy(lbuffer,buffer,len);
  lbuffer[len]=0;
  *result=simple_strtoul(lbuffer,&endptr,0)*sign;
  if(!(*endptr) || isspace(*endptr)) return 0;
  return -ERANGE;
}

static int xadio_scanhex(const char *buffer,int len,char *destination,int size,int *lenPtr) {
int i,currVal=0,pos=0;
  *lenPtr=0;
  for(i=0;i<len;i++) {
    int val=buffer[i];
    if(isspace(val)) continue;
    switch(val) {
      case '0'...'9': val-='0'; break;
      case 'A'...'F': val-='A'-10; break;
      case 'a'...'f': val-='a'-10; break;
      default:
        return -EINVAL;
    }
    if(i&1) {
      currVal|=(val&0xf)<<4;
      if(pos<size) {
        destination[pos++]=currVal;
      }
    } else {
      currVal=val&0xf;
    }
  }
  if(i&1) {
    return -EINVAL;
  }
  *lenPtr=pos;
  return 0;
}

static void control_interpreter(int minor,const char *buffer,int len) {
#define TN(x) (len>=(int)strlen(x) && !strncmp(buffer,x,strlen(x))) 
  if(!len) return;
  if(TN("setfreq")) {
    int i,value;
    for(i=0;i<len;i++) if(isspace(buffer[i])) break;
    for(   ;i<len;i++) if(!isspace(buffer[i])) break;
    if(xadio_scanint(&buffer[i],len-i,&value)) goto error;
    if(xadio_setfreq(minor,value)) goto error;
  } else if(TN("setincmode0")) {
    int i,value;
    for(i=0;i<len;i++) if(isspace(buffer[i])) break;
    for(   ;i<len;i++) if(!isspace(buffer[i])) break;
    if(xadio_scanint(&buffer[i],len-i,&value)) goto error;
    if(value<0 || value>3) goto error;
    xadio_setincmode0(minor,value);
  } else if(TN("setincmode1")) {
    int i,value;
    for(i=0;i<len;i++) if(isspace(buffer[i])) break;
    for(   ;i<len;i++) if(!isspace(buffer[i])) break;
    if(xadio_scanint(&buffer[i],len-i,&value)) goto error;
    if(value<0 || value>3) goto error;
    xadio_setincmode1(minor,value);
  } else if(TN("resetcounter")) {
  } else if(TN("calibrate")) {
  } else if(TN("resetadu")) {
  } else if(TN("resetint")) {
  } else if(TN("dmastop")) {
  } else if(TN("dmastart")) {
  } else if(TN("reset")) {
  } else if(TN("setdigio")) {
    int i,value,mask;
    for(i=0;i<len;i++) if(isspace(buffer[i])) break;
    for(   ;i<len;i++) if(!isspace(buffer[i])) break;
    if(xadio_scanint(&buffer[i],len-i,&value)) goto error;
    for(   ;i<len;i++) if(isspace(buffer[i])) break;
    for(   ;i<len;i++) if(!isspace(buffer[i])) break;
    if(xadio_scanint(&buffer[i],len-i,&mask)) {
      mask=0xfffffff;
    }
    xadio_setdigio(minor,value,mask);
  } else if(TN("setdax0")) {
    int i,value;
    for(i=0;i<len;i++) if(isspace(buffer[i])) break;
    for(   ;i<len;i++) if(!isspace(buffer[i])) break;
    if(xadio_scanint(&buffer[i],len-i,&value)) goto error;
    xadio_setdac(minor,0,value);
  } else if(TN("setdax1")) {
    int i,value;
    for(i=0;i<len;i++) if(isspace(buffer[i])) break;
    for(   ;i<len;i++) if(!isspace(buffer[i])) break;
    if(xadio_scanint(&buffer[i],len-i,&value)) goto error;
    xadio_setdac(minor,1,value);
  } else if(TN("setda0")) {
    int i,value;
    for(i=0;i<len;i++) if(isspace(buffer[i])) break;
    for(   ;i<len;i++) if(!isspace(buffer[i])) break;
    if(xadio_scanint(&buffer[i],len-i,&value)) goto error;
    if(value<-10240000 || value>10240000) goto error;
    xadio_setdacuv(minor,0,value);
  } else if(TN("setda1")) {
    int i,value;
    for(i=0;i<len;i++) if(isspace(buffer[i])) break;
    for(   ;i<len;i++) if(!isspace(buffer[i])) break;
    if(xadio_scanint(&buffer[i],len-i,&value)) goto error;
    if(value<-10240000 || value>10240000) goto error;
    xadio_setdacuv(minor,1,value);
  } else if(TN(XADIO_CONTROL_setconfigrom)) {
    int i,res;
    char lbuffer[XADIO_CONFIG_ROMSIZE];
    for(i=0;i<len;i++) if(isspace(buffer[i])) break;
    for(   ;i<len;i++) if(!isspace(buffer[i])) break;
    if(xadio_scanhex(&buffer[i],len-i,lbuffer,sizeof(lbuffer),&res)) goto error;
    memcpy(xadio_table[minor].config_rom,lbuffer,res);
  } else {
    goto error;
  }
  return;
error:
  printk(KERN_WARNING "%s error in control device '%.20s`\n",XADIO_NAME(minor), buffer);
#undef TN
}

static int xadio_control_write(struct file *file, const char *buffer, unsigned long count, void *data) {
unsigned int minor=((unsigned int)data) & 0xff;
loff_t off=file->f_pos;
int lstart=0;
int i;
  if(debug>1) printk(KERN_DEBUG "%s control_write: off:%d count:%d value:%s",XADIO_NAME(minor),(int)off,(int)count,buffer);
  for(i=0;i<(int)count;i++) {
    if(buffer[i]=='\n') {
      control_interpreter(minor,buffer+lstart,i-lstart);
      lstart=i+1;
    }
  }
  if(lstart!=(int)count) {
    control_interpreter(minor,buffer+lstart,i-lstart);
  }
  return count;
}

static struct proc_dir_entry *xadioroot;

static void xadio_proc_del_tree(void) {
struct proc_dir_entry *p;
  if(!xadioroot) return;
  for(p=xadioroot->subdir;p;) {
    struct proc_dir_entry *q;
    for(q=p->subdir;q;) {
      struct proc_dir_entry *r=q->next;
      remove_proc_entry(q->name,p);
      q=r;
    }
    q=p->next;
    remove_proc_entry(p->name,xadioroot);
    p=q;
  }
  remove_proc_entry(XADIO_STRING,0);
}

static void xadio_proc_register(unsigned int minor) {
struct proc_dir_entry *card,*entry;
  if(!xadioroot) {
    if(!(xadioroot=create_proc_entry(XADIO_STRING ,S_IFDIR|0555,0))) return;
  }
  if(!(card=create_proc_entry(XADIO_NAME(minor),S_IFDIR|0555,xadioroot))) return;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
  if(!(entry = proc_mknod("samples",S_IFCHR|0666,card,MKDEV(major,minor)))) return;
#else
  if(!(entry=create_proc_entry("samples",S_IFCHR|0666,card))) return;
  entry->proc_fops=&xadio_fops;
  entry->data=(void*)(minor & 0xff);
#endif
  //entry->rdev=MKDEV(major,minor);
  entry->proc_fops=&xadio_fops;
  if(!(entry=create_proc_entry("control",S_IFREG|0666,card))) return;
  entry->read_proc=xadio_control_read;
  entry->write_proc=xadio_control_write;
  entry->data=(void*)(minor & 0xff);
  if(!(entry=create_proc_entry("values",S_IFREG|0666,card))) return;
  entry->read_proc=xadio_values_read;
  entry->write_proc=xadio_control_write;
  entry->data=(void*)(minor & 0xff);
}

static int __devinit xadio_initcard(int minor) {
  if(minor<0 || minor>1) return -ENOSPC;
  sprintf(xadio_table[minor].name,XADIO_STRING "%d",minor);
  sprintf(xadio_table[minor].config_rom,XADIO_STRING "%d 9axkljnaslkdj908234jklsnmsdfaml8HJSHD",minor);
  xadio_table[minor].rate=1;
  printk(KERN_DEBUG "%s card simulated:\n",XADIO_NAME(minor));
  xadio_proc_register(minor);
  return 0;
}

static int __init xadio_init(void) {
  bsize=brecords*sizeof(Xadio_record);
  xadio_initcard(0);
  xadio_initcard(1);
  register_chrdev(major, XADIO_STRING ,&xadio_fops);
  printk(KERN_INFO XADIO_STRING " driver loaded\n");
  return 0;
}

static void __exit xadio_exit(void) {
  unregister_chrdev(major, XADIO_STRING);
  xadio_proc_del_tree();
  printk(KERN_INFO XADIO_STRING " driver unloaded\n");
}

/* ------------------- Module-Code  ------------------ */

MODULE_AUTHOR("Peter Brueckner (peter@bj-ig.de)");
MODULE_DESCRIPTION("Simulation-Driver for XADIO Cards");
MODULE_SUPPORTED_DEVICE("XADIO XTREME AnalogDigitalIO 8");
MODULE_LICENSE("GPL v2");

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18))
MODULE_PARM(debug, "i");
MODULE_PARM(major, "i");
MODULE_PARM(brecords,"i");
MODULE_PARM(bcount,"i");
#else
module_param(debug, int,0600);
module_param(major, int,0600);
module_param(brecords,int,0600);
module_param(bcount,int,0600);
#endif

MODULE_PARM_DESC(debug,"Debuglevel (default:0)");
MODULE_PARM_DESC(major,"Major device number (default:0x46/70)");
MODULE_PARM_DESC(brecords,"Buffersize in Samples (64 Bytes) (default:"__MODULE_STRING(XADIO_DEFAULT_BRECORDS)" Records)");
MODULE_PARM_DESC(bcount,"Buffercount per card 2..16(default:" __MODULE_STRING(XADIO_DEFAULT_BCOUNT) ")");

module_init(xadio_init);
module_exit(xadio_exit);
